(**
 ** $Id: Complex.sml 89 2005-12-16 17:00:43Z rlpm $
 **
 ** A complex number library for Standard ML
 ** Author: Rory McGuire <rlpm at cs dot unm dot edu>
 **
 ** This program is free software; you can redistribute it and/or
 ** modify it under the terms of the GNU General Public License as
 ** published by the Free Software Foundation; either version 2 of the
 ** License, or (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 ** General Public License for more details.
 **
 ** You can obtain a copy of the GPL at:
 **   http://www.fsf.org/copyleft/gpl.html.
 **)

structure Complex : COMPLEX =
struct
  infix 6 + -
  infix 7 * /

  type complex = {real:real, imag:real}

  val i = {real=0.0,imag=1.0}

  fun fromReal r =
      {real=r,imag=0.0}

  val fromInt = fromReal o Real.fromInt

  fun magnitude {real,imag} =
      Math.sqrt(Real.+(Real.*(real,real),Real.*(imag,imag)))

  fun phase {real,imag} =
      if Real.==(imag,0.0)
      then if Real.signBit real
           then Math.pi
           else 0.0
      else if Real.==(real,0.0)
      then if Real.signBit imag
           then Real.~(Real./(Math.pi,2.0))
           else Real./(Math.pi,2.0)
      else
          let
              val a = Math.atan (Real./(imag,real))
          in
              if Real.signBit real
              then if Real.signBit imag
                   then a - Math.pi
                   else a + Math.pi
              else a
          end

  fun toPolar x =
      {magnitude=magnitude x, phase=phase x}

  fun fromPolar {magnitude,phase} =
      {real=Real.*(magnitude,Math.cos phase),
       imag=Real.*(magnitude,Math.sin phase)}

  fun toString {real,imag} =
      Real.toString real ^ "+" ^ Real.toString imag ^ "i"

  fun ~ {real,imag} =
      {real=Real.~(real),imag=Real.~(imag)}

  fun conj {real,imag} =
      {real=real,imag=Real.~(imag)}

  val ~~ = conj

  fun op+ ({real=r1,imag=i1},{real=r2,imag=i2}) =
      {real=Real.+(r1,r2),imag=Real.+(i1,i2)}
  
  fun op- ({real=r1,imag=i1},{real=r2,imag=i2}) =
      {real=Real.-(r1,r2),imag=Real.-(i1,i2)}

  fun op* ({real=x,imag=y},{real=a,imag=b}) =
      {real=Real.-(Real.*(x,a),Real.*(y,b)),
       imag=Real.+(Real.*(y,a),Real.*(x,b))}
  
  fun op/ ({real=x,imag=y},{real=a,imag=b}) =
      let
          fun f x = Real./(x,Real.+(Real.*(a,a),Real.*(b,b)))
      in
          {real=f (Real.+(Real.*(x,a),Real.*(y,b))),
           imag=f (Real.+(Real.*(y,a),Real.*(x,b)))}
      end
  

  fun exp {real,imag} =
      fromPolar{magnitude=Math.exp(real),phase=imag}
 
  fun ln x =
      let
          val {magnitude,phase} = toPolar x
      in
          {real=Math.ln magnitude,imag=phase}
      end
 
  fun pow (z, {real=x,imag=y}) =
      let
          val {magnitude=a,phase=p} = toPolar z
          val lna = Math.ln a
      in
          exp({real=Real.-(Real.*(x,lna),Real.*(p,y)),
               imag=Real.+(Real.*(p,x),Real.*(y,lna))})
      end

  fun sin {real=x,imag=y} =
      {real=Real.*(Math.sin x,Math.cosh y),
       imag=Real.*(Math.cos x,Math.sinh y)}

  fun cos {real=x,imag=y} =
      {real=Real.*(Math.cos x,Math.cosh y),
              imag=Real.~(Real.*(Math.sin x,Math.sinh y))}
end
