(**
 ** $Id: Fourier.sml 90 2005-12-16 17:01:35Z rlpm $
 ** 
 ** A Discrete Fourier Transform library for Standard ML
 ** Author: Rory McGuire <rlpm at cs dot unm dot edu>
 **
 ** This program is free software; you can redistribute it and/or
 ** modify it under the terms of the GNU General Public License as
 ** published by the Free Software Foundation; either version 2 of the
 ** License, or (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 ** General Public License for more details.
 **
 ** You can obtain a copy of the GPL at:
 **   http://www.fsf.org/copyleft/gpl.html.
 **)
structure Fourier : FOURIER =
struct
  val twoPi = Complex.*(Complex.fromReal 2.0, Complex.fromReal Math.pi)

  fun dftHelper inv f =
      let
          val bigN = Vector.length f
          val bigN_complex = Complex.fromInt bigN
          val sqrtN = Complex.pow(Complex.fromReal (Real.fromInt bigN),
                                  Complex.fromReal 0.5)
          val ti = if inv then Complex.~ Complex.i else Complex.i
          fun inner m =
              let
                  fun helper n =
                      Complex.*(Vector.sub(f,n),
                                Complex.exp
                                    (Complex./
                                         (Complex.*
                                              (Complex.*(ti,twoPi),
                                               Complex.*(Complex.fromInt n,
                                                         Complex.fromInt m)),
                                          bigN_complex)))
              in
                  Complex./
                      (Vector.foldl
                           Complex.+
                           (Complex.fromReal 0.0)
                           (Vector.tabulate(bigN,helper)),
                       sqrtN)
              end
      in
          Vector.tabulate(bigN,inner)
      end

  val dft = dftHelper false
  val idft = dftHelper true

  fun twoDimHelper f a =
      let
          fun rowMap a =
              let
                  val v = Vector.tabulate
                              (Array2.nRows a,
                               (fn x => f (Array2.row(a,x))))
              in
                  Array2.tabulate
                      Array2.RowMajor
                      (Array2.nRows a,
                       Array2.nCols a,
                       (fn (i,j) => Vector.sub(Vector.sub(v,i),j)))
              end
          fun transpose a =
              Array2.tabulate
                  Array2.RowMajor
                  (Array2.nCols a,
                   Array2.nRows a,
                   (fn (j,i) => Array2.sub(a,i,j)))
      in
          (transpose o rowMap o transpose o rowMap) a
      end

  val dft2d = twoDimHelper (dftHelper false)
  val idft2d = twoDimHelper (dftHelper true)

  (* Fast Fourier Transform
   * http://en.wikipedia.org/wiki/Cooley-Tukey_FFT_algorithm
   *)

  fun fftHelper inv v =
      let
          val n = Vector.length v
      in
          if n = 1 then v else
          let
              val ti = if inv then Complex.~ Complex.i else Complex.i
              fun vref x = Vector.sub(v,x)
              val n_2 = n div 2
              val fe = fftHelper inv
                                 (Vector.tabulate
                                      (n_2,(fn x => vref(2 * x))))
              val fo = fftHelper inv
                                 (Vector.tabulate
                                      (n_2,(fn x => vref(2 * x + 1))))
              fun feref x = Vector.sub(fe,x)
              fun foref x = Vector.sub(fo,x)
          in
              Vector.tabulate
                  (n,
                   (fn j =>
                       if j < n_2
                       then
                           Complex.+(feref j,
                                     Complex.*
                                         ((Complex.exp o Complex.* )
                                              (Complex.*(ti,twoPi),
                                               Complex./(Complex.fromInt j,
                                                         Complex.fromInt n)),
                                          foref j))
                       else
                           let
                               val j_n_2 = j - n_2
                           in
                               Complex.-(feref j_n_2,
                                         Complex.*
                                             ((Complex.exp o Complex.* )
                                                  (Complex.*(ti,twoPi),
                                                   Complex./
                                                       (Complex.fromInt j_n_2,
                                                        Complex.fromInt n)),
                                              foref j_n_2))
                           end))
          end
      end
      
      fun isTwoPower 0 = true
        | isTwoPower 1 = true
        | isTwoPower x =
          let
              val y = x div 2
          in
              x = y * 2 andalso isTwoPower y
          end

      fun fftSizeCheck v =
          if isTwoPower (Vector.length v)
          then true
          else raise Fail "FFT needs power-of-two size"

      fun fftHelper2 inv v =
          let
              val _ = fftSizeCheck v
              val sq = (Complex.fromReal o Math.sqrt o
                        Real.fromInt o Vector.length) v
          in
              Vector.map (fn x => Complex./(x,sq)) (fftHelper inv v)
          end

      val fft = fftHelper2 false
      val ifft = fftHelper2 true                
      val fft2d = twoDimHelper (fftHelper2 false)
      val ifft2d = twoDimHelper (fftHelper2 true)
end
